/*
 * LVRM.cc
 * - load-aware virtual router monitor handles issues among VRs e.g. core allocation
 */
#include "LVRM.hh"

/*
 * Constructor
 */
LVRM::LVRM() {	// handle interrupt from keyboard and invalid memory reference
	signal(SIGINT, handler);
	signal(SIGSEGV, handler);
}

/*
 * Destructor
 */
LVRM::~LVRM() {	// deallocate all VR instance monitors
	for (int i = 0; i < vrInstanceMonitorCount; i++) {
		delete vrInstanceMonitor[i];
	}
}

/*
 * Allocate a VR instance monitor of the VR name specified by the parameters
 */
VRInstanceMonitor* LVRM::createVRInstanceMonitor(KeyToVRIM* keyToVRIM, int keyCount, char* vrName) {
	vrInstanceMonitor = (VRInstanceMonitor**) realloc(vrInstanceMonitor, ++vrInstanceMonitorCount * sizeof(VRInstanceMonitor*));
	vrInstanceMonitor[vrInstanceMonitorCount - 1] = new VRInstanceMonitor(vrName);

	for (int i = 0; i < keyCount; i++) {	// append keys to the list of keys to VR instance monitors
		keyToVRIM[i].vrInstanceMonitor = vrInstanceMonitor[vrInstanceMonitorCount - 1];
		createKeyToVRIM(keyToVRIM + i);
	}

	return vrInstanceMonitor[vrInstanceMonitorCount - 1];	// return the VR instance monitor
}

/*
 * Initialize the LVRM to allow dynamically allocating the core allocator and the socket adapter on the heap
 */
int LVRM::initialize() {
	cpu_set_t cpuset;	// set of CPUs allocating to the LVRM
	int rtn;			// return value of the system call nice()

	if (getuid()) {	// return error if user is not root
		return -1;
	}

	rtn = nice(-20);
	CPU_ZERO(&cpuset);
	CPU_SET(CPU, &cpuset);
	coreAllocator = new CORE_ALLOCATOR(&vrInstanceMonitor, &vrInstanceMonitorCount);
	sched_setaffinity(0, sizeof(cpu_set_t), &cpuset);
	socketAdapter = new SOCKET_ADAPTER();	// "LVRM.hh" defines SOCKET_ADAPTER
	return 0;
}

/*
 * Try to receive a network frame from a VR instance monitor
 *
 * @param buffer - frame buffer inside LVRM
 */
int LVRM::fromVRInstanceMonitor(VRInstanceAdapter::Buffer* buffer) {
	int frame_len;	// actual frame size

	for (int i = 0; i < vrInstanceMonitorCount; i++) {	// return the number of bytes received from a VR instance w.r.t. the source MAC
		if ((frame_len = vrInstanceMonitor[i]->fromVRInstanceAdapter(buffer))) {
			return frame_len;
		}
	}

	return 0;	// return no byte received
}

/*
 * Send the network frame to the VR instance monitor
 *
 * @param buffer - frame buffer inside LVRM
 * @param frame_len - actual length of the frame
 */
int LVRM::toVRInstanceMonitor(VRInstanceAdapter::Buffer* buffer, int frame_len) {
	for (int i = 0; i < keyCount; i++) {	// return the number of bytes sent to the VR instance w.r.t. the source MAC address in the frame
		if (ntohl(*(int*) keyToVRIM[i].key) == (ntohl(((iphdr*) (buffer->buffer + sizeof(ethhdr)))->daddr) & -1 << (32 - keyToVRIM[i].key[IPVERSION]))) {
			coreAllocator->allocate();
			return keyToVRIM[i].vrInstanceMonitor->toVRInstanceAdapter(buffer, frame_len);
		}
	}

	return 0;	// return no byte sent
}

/*
 * Try to transceive a network frame via the LVRM
 * 
 * first, check frames sent to the network; then check frames received from the network (always give higher priority to the frames sent to
 * network)
 */
int LVRM::transceive() {
	VRInstanceAdapter::Buffer buffer;	// try to receive a network frame from the socket adapter to the LVRM
	buffer.sd = 0;
	int frame_len = fromVRInstanceMonitor(&buffer);	// actual frame size

	if (0 < frame_len) {	// return the number of bytes transceived
#ifdef PRINT_END
		((PCapFile*) socketAdapter)->end();
#endif
		return socketAdapter->fromLVRM(buffer.buffer, frame_len + ((char*) &buffer - buffer.buffer));
	} else if ((frame_len = socketAdapter->toLVRM(buffer.buffer))) {
#ifdef PRINT_END
		int rtn = toVRInstanceMonitor(&buffer, frame_len + (buffer.buffer - (char*) &buffer)); //return the actual frame size
		((PCapFile*) socketAdapter)->start();
		return rtn;
#endif
		return toVRInstanceMonitor(&buffer, frame_len + (buffer.buffer - (char*) &buffer));
	}

	return 0;	// return no byte transceived
}

/*
 * Allocate a key-to-VRIM mapping
 */
void LVRM::createKeyToVRIM(KeyToVRIM* keyToVRIM_par) {
	keyToVRIM = (KeyToVRIM*) realloc(keyToVRIM, ++keyCount * sizeof(KeyToVRIM));
	keyToVRIM[keyCount - 1] = *keyToVRIM_par;
}

/*
 * Try to deallocate the key-to-VRIM mapping
 *
 * @param keyToVRIM_par - the mapping to be deleted
 */
void LVRM::destroyKeyToVRIM(KeyToVRIM* keyToVRIM_par) {
	for (int i = 0; i < keyCount; i++) {
		if (!memcmp(keyToVRIM + i, keyToVRIM_par, sizeof(KeyToVRIM))) {
			memcpy(keyToVRIM + i, keyToVRIM + i + 1, (--keyCount - i) * sizeof(KeyToVRIM));
		}
	}

	keyToVRIM = (KeyToVRIM*) realloc(keyToVRIM, keyCount * sizeof(KeyToVRIM));	// resize the list
}

/*
 * Try to deallocate the VR instance monitor pointed by the parameter
 *
 * @param vrInstanceMonitor_par - the VR to be deleted
 */
void LVRM::destroyVRInstanceMonitor(VRInstanceMonitor* vrInstanceMonitor_par) {
	for (int i = 0; i < vrInstanceMonitorCount; i++) {	// remove the VR from the list of VRs if the parameter is in the list
		if (vrInstanceMonitor[i] == vrInstanceMonitor_par) {
			for (int j = 0; j < keyCount; j++) {
				if (vrInstanceMonitor[i] == keyToVRIM[j].vrInstanceMonitor) {
					memcpy(keyToVRIM + j, keyToVRIM + j + 1, (--keyCount - j) * sizeof(KeyToVRIM));
				}
			}

			keyToVRIM = (KeyToVRIM*) realloc(keyToVRIM, keyCount * sizeof(KeyToVRIM));	// Remove the given VR and resize the VR list
			delete vrInstanceMonitor[i];
			memcpy(vrInstanceMonitor + i, vrInstanceMonitor + i + 1, (--vrInstanceMonitorCount - i) * sizeof(VRInstanceMonitor*));
			vrInstanceMonitor = (VRInstanceMonitor**) realloc(vrInstanceMonitor, vrInstanceMonitorCount * sizeof(VRInstanceMonitor*));
		}
	}
}

/*
 * ANSI C signal handling
 */
void LVRM::handler(int) { // deallocate lvrm and successfully terminate current LVRM process
	delete lvrm;
	exit(0);
}
