/*
 * if.cc
 * - map Ethernet interface name to MAC address, or vice versa, based on the
 * kernel setting.
 */
#include <net/if.h>
#include "if.hh"

/*
 * map a network interface name to its corresponding MAC
 * 
 * @param ifmac - network interface name
 * @return -1 - if error, 0 if success
 */
int if_nametomac(char *ifname, __u8 *ifmac) {	// map a network interface name to its corresponding MAC address
	int nSD;									// socket descriptor
	struct if_nameindex *pListSave;				// ptr to interface name index
	struct ifreq sIfReq;						// interface request

#ifndef SIOCGIFADDR
	return -1;	// the kernel does not support the required ioctl()s
#endif

	if ((nSD = socket(PF_INET, SOCK_STREAM, 0)) < 0) {						// create a socket that we can use for all of our ioctls
		printf("File %s: line %d: Socket failed\n", __FILE__, __LINE__);	// socket creation failed, this is a fatal error
		return -1;
	}

	// obtain a list of dynamically allocated structures
	struct if_nameindex *pIfList = pListSave = if_nameindex();	// ptr to interface name index

	for (; *(char *) pIfList; pIfList++) {		// walk thru the array returned and query for each interface's address
		if (strcmp(pIfList->if_name, ifname)) {	// determine if we are processing the interface that we are interested in
			continue;							// nope, check the next one in the list
		}

		strncpy(sIfReq.ifr_name, pIfList->if_name, IF_NAMESIZE);

		if (ioctl(nSD, SIOCGIFHWADDR, &sIfReq)) {							// get the MAC address for this interface
			printf("File %s: line %d: Ioctl failed\n", __FILE__, __LINE__);	// we failed to get the MAC address for the interface
			return -1;
		}

		memmove(ifmac, &sIfReq.ifr_ifru.ifru_hwaddr.sa_data[0], 6);
		break;
	}

	if_freenameindex(pListSave);	// clean up things and return
	close(nSD);
	return 0;
}

/*
 * map a network interface name to its corresponding index
 */
unsigned nf_nametoindex(const char *ifname) {
	return if_nametoindex(ifname);	// wrapper function, map a network interface name to its corresponding index
}
