/*
 * VRInstanceAdapter.cc
 * - connect to the VR instance
 */

#include "VRInstanceAdapter.hh"

/*
 * Destructor
 */
VRInstanceAdapter::~VRInstanceAdapter() {	// send interrupt from keyboard to the VR instance and deallocate shared memory
	kill(pid, SIGINT);
	shmdt(shmaddr);
	shmctl(shmid, IPC_RMID, NULL);
	wait(0);
}

/*
 * Try to receive a network frame from the VR instance
 */
int VRInstanceAdapter::fromVRInstance(Buffer *buffer) {
	int len = ipcBuffer[CTRL_TO_LVRM]->Extract(buffer);	// try to receive a control message from the control ring buffer

	if (0 < len) {	// if a control message is ready, return the number of bytes received from the control ring buffer
		return len;
	}

	return ipcBuffer[DATA_TO_LVRM]->Extract(buffer);	// return the number of bytes received from the data ring buffer
}

/*
 * Get the ID of the CPU dedicated
 */
int VRInstanceAdapter::getCpu() {
	return cpu;
}

/*
 * Initialize
 */
void VRInstanceAdapter::initialize(char* vrName, int cpu_par) {
	char str[17];							// shmid string
	cpu_set_t cpuset;						// set of CPUs allocating to the VR instance
	RingBuffer<Buffer> src(RB_SIZE, false);	// template of IPC buffers

	struct {
		int buffer_len;											// length of ring buffer
		int control_len;										// length of ring buffer excluding data part
		RingBuffer<Buffer>* src;								// source of ring buffer
	} bufferInit[] = {{SHMMAX / 4, 5 * CACHE_LINE_SIZE, &src},	// ring buffer #0
		{SHMMAX / 4, 5 * CACHE_LINE_SIZE, &src},				// ring buffer #1
		{SHMMAX / 4, 5 * CACHE_LINE_SIZE, &src},				// ring buffer #2
		{SHMMAX / 4, 5 * CACHE_LINE_SIZE, &src}};				// ring buffer #3

	CPU_ZERO(&cpuset);
	CPU_SET(cpu = cpu_par, &cpuset);
	latencyEstimator = new LATENCY_ESTIMATOR(this);
	shmid = shmget(IPC_PRIVATE, SHMMAX, IPC_CREAT); // allocate and attach to the shared memory segment
	sprintf(str, "shmid=%d", shmid);
	ipcBuffer = (RingBuffer<Buffer>**) malloc(4 * sizeof(RingBuffer<Buffer>*));
	ipcBuffer[0] = (RingBuffer<Buffer>*) (shmaddr = (char *) shmat(shmid, 0, SHM_RND));
	memcpy(ipcBuffer[0], bufferInit[0].src, bufferInit[0].control_len);

	for (int i = 1; i < 4; i++) {	// initialize other ring buffers
		ipcBuffer[i] = (RingBuffer<Buffer>*) ((char*) ipcBuffer[i - 1] + bufferInit[i - 1].buffer_len);
		memcpy(ipcBuffer[i], bufferInit[i].src, bufferInit[i].control_len);
	}

	if (!(pid = vfork())) {	// create a child process to execute the VR with shmid and block parent
		sched_setaffinity(0, sizeof(cpu_set_t), &cpuset);
		execlp(vrName, vrName, str, NULL);
		execlp("click", "click", vrName, str, NULL);
	}
}

/*
 * Send a network frame to the VR instance
 */
int VRInstanceAdapter::toVRInstance(Buffer *buffer, int buffer_len) {
	if (buffer->sd) {	// if it is from another VR instance, return the number of bytes sent to the control ring buffer
		return ipcBuffer[CTRL_TO_VR_INSTANCE]->Insert(*buffer, buffer_len);
	}

	return ipcBuffer[DATA_TO_VR_INSTANCE]->Insert(*buffer, buffer_len);	// return the number of bytes sent to data ring buffer
}
