/*
 * VRInstanceMonitor.cc
 * - VR instance monitor handles issues within the VR e.g. load balance
 */
#include "VRInstanceMonitor.hh"

/*
 * Constructor
 */
VRInstanceMonitor::VRInstanceMonitor(char* vrName) : vrInstanceAdapter(0), vrName(vrName), vrInstanceAdapterCount(0) {
	loadBalancer = new LOAD_BALANCER(&vrInstanceAdapterCount, &vrInstanceAdapter);	// initialize some variables
	throughputEstimator = new THROUGHPUT_ESTIMATOR();
}

/*
 * Destructor
 */
VRInstanceMonitor::~VRInstanceMonitor() {	// deallocate all VR instances
	for (int i = 0; i < vrInstanceAdapterCount; i++) {
		delete vrInstanceAdapter[i];
	}
}

/*
 * return the throughput
 */
float VRInstanceMonitor::throughput() {
	return CPU_HZ / throughputEstimator->getEstimate();
}

/*
 * try to receive a network frame from a VR instance adapter to the VR instance monitor
 */
int VRInstanceMonitor::fromVRInstanceAdapter(VRInstanceAdapter::Buffer* buffer) {
	int len;	// length of network frame to be received

	for (int i = 0; i < vrInstanceAdapterCount; i++) {	// search for a receivable network frame and return its length
		if (0 < (len = vrInstanceAdapter[i]->fromVRInstance(buffer))) {
			if (buffer->sd) {
				vrInstanceAdapter[buffer->sd - 1]->toVRInstance(buffer, len);
			} else {
				return len;
			}
		}
	}

	return 0;	// return no byte received
}

/*
 * return the number of VR instance adapters
 */
int VRInstanceMonitor::getVRInstanceAdapterCount() {
	return vrInstanceAdapterCount;
}

/*
 * Send a network frame from the VR instance monitor to the VR instance adapter
 */
int VRInstanceMonitor::toVRInstanceAdapter(VRInstanceAdapter::Buffer* buffer, int frame_len) {	// return the number of bytes sent
	throughputEstimator->estimate();	// sample the throughput of the VR instance
	return vrInstanceAdapter[loadBalancer->balance()]->toVRInstance(buffer, frame_len);
}

/*
 * Allocate a VR instance adapter
 */
void VRInstanceMonitor::createVRInstanceAdapter(int cpu) {	// append a VR instance adapter to the list of VR instance adapters
	vrInstanceAdapter = (VRInstanceAdapter**) realloc(vrInstanceAdapter, ++vrInstanceAdapterCount * sizeof(VRInstanceAdapter*));
	vrInstanceAdapter[vrInstanceAdapterCount - 1] = new VR_INSTANCE_ADAPTER();
	vrInstanceAdapter[vrInstanceAdapterCount - 1]->initialize(vrName, cpu);
}

/*
 * Remove a given CPU from a set
 */
void VRInstanceMonitor::destroyVRInstanceAdapter(int cpu) {
	for (int i = 0; i < vrInstanceAdapterCount; i++) {
		if (cpu == vrInstanceAdapter[i]->getCpu()) {
			delete vrInstanceAdapter[i];
			memcpy(vrInstanceAdapter + i, vrInstanceAdapter + i + 1, (--vrInstanceAdapterCount - i) * sizeof(VRInstanceAdapter*));
		}
	}

	vrInstanceAdapter = (VRInstanceAdapter**) realloc(vrInstanceAdapter, vrInstanceAdapterCount * sizeof(VRInstanceAdapter*));
}	// Remove the given CPU from the CPU list and resize the CPU list
